from cooptools.matrixManipulation import point_transform_3d, scaled_array
import pygame
import numpy as np
from typing import List, Dict, Tuple, Callable
from dataclasses import dataclass
from cooptools.colors import Color
import cooptools.geometry_utils.vector_utils as vec

ColorGetter = Callable[[int], Color]

@dataclass(frozen=True)
class DrawPointArgs:
    color: Color | ColorGetter = None
    outline_color: Color | ColorGetter = None
    radius: int = None
    outline_width: int = None

    def get_color(self, ii) -> Color:
        if type(self.color) == ColorGetter:
            return self.color(ii)
        return self.color

    def get_outline_color(self, ii) -> Color:
        if type(self.outline_color) == ColorGetter:
            return self.outline_color(ii)
        return self.outline_color

def draw_points(points: Dict[Tuple[int, int], DrawPointArgs],
                surface: pygame.Surface,
                vec_transformer_getter: vec.VecTransformerGetter = None):

    pt_args = [(k, v) for k, v in points.items()]

    # update the points based on the arg transformer
    scaled = vec.resolve_vec_transformed_points(
        points=[x[0] for x in pt_args],
    vec_transformer_getter=vec_transformer_getter)

    for ii, x in enumerate(pt_args):
        args = x[1]
        if args.color:


            pygame.draw.circle(surface,
                               args.get_color(ii).value,
                               center=tuple(scaled[ii][:2]),
                               radius=args.radius,
                               )

        if args.outline_color:
            olw = args.outline_width if args.outline_width is not None else 1
            pygame.draw.circle(surface,
                               args.get_outline_color(ii).value,
                               center=tuple(scaled[ii][:2]),
                               width=olw,
                               radius=args.radius,
                               )


from cooptools.coopEnum import CoopEnum
from coopstructs.geometry.vectors.vectorN import Vector2

class TextAlignmentType(CoopEnum):
    TOPLEFT = 'topleft'
    TOPRIGHT = 'topright'
    TOPCENTER = 'midtop'
    BOTTOMLEFT = 'bottomleft'
    BOTTOMRIGHT = 'bottomright'
    RIGHTCENTER = 'midright'
    BOTTOMCENTER = 'midbottom'
    LEFTCENTER = 'midleft'
    CENTER = 'center'

class PyLabel:
    def __init__(self, font, text, color, position, alignment: TextAlignmentType = TextAlignmentType.TOPLEFT):

        self.alignment = alignment
        self.image = font.render(text, 1, color)
        self.rect = self.image.get_rect()
        self.pos = position

    def draw(self, surface):
        surface.blit(self.image, self.rect)

    def set_alignment(self, alignment: TextAlignmentType, pos: Vector2 = None):
        self.alignment = alignment
        self.pos = pos
        setattr(self.rect, self.alignment.value, self.pos.as_tuple())


import harfang as hg
from coopstructs.vectors import Vector2

def draw_circle(plus):
    # plus.Circle2D()
    draw_triangle(plus)

def draw_square(plus, pos: Vector2, size: int):
    cube = plus.CreateGeometry(plus.CreateCube())
    plus.Geometry2D(int(pos.x), int(pos.y), cube, 0, 0, 0, size)

def draw_triangle(plus, a: Vector2, b: Vector2, c: Vector2, hcolor: hg.Color):
    # plus.Triangle2D(a.x, a.y, b.x, b.y, c.x, c.y, hcolor, hcolor, hcolor)

    x_adj = 0
    y_adj = 0

    plus.Triangle2D(x_adj + int(a.x), y_adj + int(a.y), x_adj + int(c.x), y_adj + int(c.y), x_adj + int(b.x), y_adj + int(b.y), hcolor, hcolor, hcolor)

    # plus.Triangle2D(int(a.x) + x_adj, int(a.y) + y_adj, int(b.x) + x_adj, int(b.y) + y_adj, int(c.x) + x_adj, int(c.y) + y_adj,
    #                 hg.Color.Red, hg.Color.Blue, hg.Color.Green)
    # plus.Triangle2D(40, 40, 200, 260, 360, 40, hg.Color.Red, hg.Color.Blue, hg.Color.Green)


import pygame
from coopstructs.geometry.vectors.vectorN import Vector2
from coopgame.colors import Color
from typing import Dict, Tuple, Any, Union, Callable
from enum import Enum
from coopgame.spriteHandling.spritePoseHandler import SpritePoseHandler

class SpriteOriginPosition(Enum):
    CENTER = 1
    BOTTOM_MIDDLE = 2
    BOTTOM_RIGHT = 3
    TOP_LEFT = 4


class MySprite(pygame.sprite.Sprite):
    def __init__(self,
                 id: str,
                 init_pos: Vector2,
                 width: int,
                 height: int,
                 sprite_origin_position: SpriteOriginPosition = None
                 ):
        super().__init__()

        self.id = id

        self.surf = None
        self.rect = None

        self.width = None
        self.height = None

        self.set_size(width, height)

        self.set_pos(init_pos, sprite_origin_position)

    @property
    def size(self):
        return (self.width, self.height)

    def set_size(self, width, height):
        if width != self.width or height != self.height:
            self.width = width
            self.height = height

            self.surf = pygame.Surface([width, height]).convert()
            self.surf.fill(Color.PINK.value)
            self.surf.set_colorkey(Color.PINK.value)

            self.rect = self.surf.get_rect()

    @property
    def bottom_center_pos(self) -> Vector2:
        x = self.rect.x + self.rect.width / 2
        y = self.rect.y + self.rect.height

        return Vector2(x, y)

    def _center_from_sprite_origin(self, pos: Vector2, sprite_origin_position: SpriteOriginPosition):
        if sprite_origin_position == SpriteOriginPosition.CENTER:
            ret = (pos.x, pos.y)
        elif sprite_origin_position == SpriteOriginPosition.BOTTOM_MIDDLE:
            ret = (pos.x, pos.y - self.surf.get_height() / 2)
        elif sprite_origin_position == SpriteOriginPosition.BOTTOM_RIGHT:
            ret = (pos.x - self.surf.get_width() / 2, pos.y - self.surf.get_height() / 2)
        elif sprite_origin_position == SpriteOriginPosition.TOP_LEFT:
            ret = (pos.x + self.surf.get_width() / 2, pos.y + self.surf.get_height() / 2)
        else:
            raise NotImplementedError(f"origin pos: {sprite_origin_position} not handled for finding center")
        return ret

    def set_pos(self, pos: Vector2, sprite_origin_position: SpriteOriginPosition = None):
        if sprite_origin_position is None:
            sprite_origin_position = SpriteOriginPosition.TOP_LEFT

        center = self._center_from_sprite_origin(pos, sprite_origin_position)

        self.rect = self.surf.get_rect(center=center)

    def blit(self, surface: pygame.Surface,
             display_handle: Union[bool, Callable[[], bool]] = False,
             display_rect: Union[bool, Callable[[], bool]] = False):

        surface.blit(self.surf, (self.rect.x, self.rect.y))
        if display_handle and callable(display_handle) and display_handle():
            pygame.draw.rect(surface, Color.RED.value, [self.rect.x, self.rect.y, 1, 1])
        elif display_handle and not callable(display_handle):
            pygame.draw.rect(surface, Color.RED.value, [self.rect.x, self.rect.y, 1, 1])

        if display_rect and callable(display_rect) and display_rect():
            pygame.draw.rect(surface, Color.RED.value, [self.rect.x, self.rect.y, self.rect.width, self.rect.height], 1)
        elif display_rect and not callable(display_handle):
            pygame.draw.rect(surface, Color.RED.value, [self.rect.x, self.rect.y, self.rect.width, self.rect.height], 1)

    def __repr__(self):
        return f"{self.id} {super().__repr__()}"


class RectangleSprite(MySprite):
    def __init__(self, id: str, init_pos: Vector2, color: Color, width: int, height: int):
        # Call the parent class (Sprite) constructor
        MySprite.__init__(self, id, init_pos, width, height)

        pygame.draw.rect(self.surf, color.value, [0, 0, width, height])


class ImageSprite(MySprite):
    def __init__(self,
                 id: str,
                 init_pos: Vector2,
                 width: int,
                 height: int):
        # Call the parent class (Sprite) constructor
        MySprite.__init__(self, id, init_pos, width, height)


class AnimatedPoseSprite(MySprite):
    def __init__(self,
                 id:str,
                 init_pos: Vector2,
                 pose_handler: SpritePoseHandler,
                 animation_cycle_ms: Union[int, Callable[[], int]] = None,
                 width: int = None,
                 height: int = None,
                 sprite_origin_position: SpriteOriginPosition = None,
                 loop:bool=True
                 ):
        self.pose_handler = pose_handler

        # Call the parent class (Sprite) constructor
        pose_w, pose_h = self.pose_handler.Shape
        width = pose_w if not width else width
        height = pose_h if not height else height
        MySprite.__init__(self, id, init_pos, width, height, sprite_origin_position)

        self.loop = loop
        self.animate_timer = 0
        self.animate_cycle = animation_cycle_ms if animation_cycle_ms else 100

        self.image = None
        self._set_image()

    def _set_image(self):
        self.surf.fill(Color.PINK.value)

        self.image = self.pose_handler.get_current()['surface']
        self.image = pygame.transform.scale(self.image, self.size)
        self.rect = self.image.get_rect(center=self.rect.center)
        self.surf.blit(self.image, (0, 0))

    def _resolve_animation_cycle(self):
        if type(self.animate_cycle) == int:
            return self.animate_cycle
        else:
            return self.animate_cycle()

    def animate(self,
                delta_time_ms: int):
        self.animate_timer += delta_time_ms
        if self.animate_timer > self._resolve_animation_cycle():
            self.increment_animation_phase(loop=self.loop)
            self.animate_timer = 0

    def increment_animation_phase(self, loop: bool = True):
        ''':return True if animation cycle has completed, False if still in cycle
        :param loop allows caller to specify whether or not to loop the aniimation (default True)'''
        ret = False
        if self.pose_handler.AnimationEnded and not loop:
            ret = True

        self.pose_handler.increment_animation(loop=loop)
        self._set_image()
        return ret

    def set_animation_cycle(self, animation_cycle_ms: Union[int, Callable[[], int]]):
        self.animate_cycle = animation_cycle_ms

class AnimatedSprite(MySprite):
    def __init__(self,
                 id: str,
                 init_pos: Vector2,
                 animation_dict: Dict[Any, Tuple],
                 default_animation_key=None,
                 animation_cycle_ms: Union[int, Callable[[], int]] = None,
                 width: int = None,
                 height: int = None,
                 sprite_origin_position: SpriteOriginPosition = None,
                 loop: bool = False):

        self._animation_dict = animation_dict

        self._animation_key = default_animation_key
        self._animation_index = 0

        # Call the parent class (Sprite) constructor
        first_image = next(iter(self._animation_dict.values()))[0]
        width = first_image.get_width() if not width else width
        height = first_image.get_height() if not height else height
        MySprite.__init__(self, id, init_pos, width, height, sprite_origin_position)

        self.loop = loop
        self.animate_timer = 0
        self.animate_cycle = animation_cycle_ms if animation_cycle_ms else 100

        self.image = None
        self._set_image()

    def set_animation(self, animation: str):
        if animation in self._animation_dict.keys() and animation != self.current_animation:
            self._animation_key = animation
            self._animation_index = 0

    def increment_animation_phase(self, loop: bool = True):
        ''':return True if animation cycle has completed, False if still in cycle
        :param loop allows caller to specify whether or not to loop the aniimation (default True)'''
        ret = False
        self._animation_index += 1
        if self._animation_index >= len(self._animation_dict[self._animation_key]):
            self._animation_index = 0 if loop else len(self._animation_dict[self._animation_key]) - 1
            ret = True

        self._set_image()
        return ret

    def _set_image(self):
        self.surf.fill(Color.PINK.value)
        self.image = self._animation_dict[self._animation_key][self._animation_index]
        self.image = pygame.transform.scale(self.image, self.size)
        self.rect = self.image.get_rect(center=self.rect.center)
        self.surf.blit(self.image, (0, 0))

    @property
    def current_animation(self):
        return self._animation_key

    def _resolve_animation_cycle(self):
        if type(self.animate_cycle) == int:
            return self.animate_cycle
        else:
            return self.animate_cycle()

    def animate(self,
                delta_time_ms: int):
        self.animate_timer += delta_time_ms
        if self.animate_timer > self._resolve_animation_cycle():
            self.increment_animation_phase(self.loop)
            self.animate_timer = 0




from cooptools.matrixManipulation import point_transform_3d, scaled_array
from coopstructs.geometry import Line
import pygame
import numpy as np
from typing import List, Dict
from dataclasses import dataclass, asdict
from cooptools.colors import Color
import coopgame.pointdrawing.point_draw_utils as putils
import coopgame.label_drawing.label_drawing_utils as lutils
from cooptools.coopEnum import Directionality
import coopgame.pygamehelpers as help
from cooptools.coopEnum import CoopEnum, auto
from cooptools.common import property_name
import cooptools.matrixManipulation as mm
import cooptools.geometry_utils.vector_utils as vec

class DirectionalityIndicatorType(CoopEnum):
    ALONG = auto()
    END = auto()

@dataclass(frozen=True)
class DirectionalityIndicatorDrawArgs:
    color: Color = None
    height: int = 10
    width: int = 10
    direction: Directionality = None

@dataclass(frozen=True)
class DrawLineArgs:
    color: Color = None
    width: int = None
    control_point_args: putils.DrawPointArgs = None
    directionality_draw_args: DirectionalityIndicatorDrawArgs = None
    draw_label_args: lutils.DrawLabelArgs = None
    label_text: str = None

    def __post_init__(self):
        if type(self.draw_label_args) == dict:
            object.__setattr__(self, 'draw_label_args', lutils.DrawLabelArgs(**self.draw_label_args))
        if type(self.control_point_args) == dict:
            object.__setattr__(self, 'control_point_args', putils.DrawPointArgs(**self.control_point_args))
        if type(self.directionality_draw_args) == dict:
            object.__setattr__(self, 'directionality_draw_args', DirectionalityIndicatorDrawArgs(**self.directionality_draw_args))

    def with_(self, **kwargs):
        kw = asdict(self)
        kw.update(kwargs)
        return DrawLineArgs(**kw)

    @property
    def BaseArgs(self):
        return DrawLineArgs(
            color=self.color,
            width=self.width,
            directionality_draw_args = self.directionality_draw_args,
        )

    @property
    def LabelArgs(self):
        return DrawLineArgs(
            draw_label_args=self.draw_label_args
        )

    @property
    def OverlayArgs(self):
        return DrawLineArgs(
            control_point_args=self.control_point_args
        )

def draw_lines(lines: Dict[Line, DrawLineArgs],
               surface: pygame.Surface,
               vec_transformer_getter: vec.VecTransformerGetter = None):
    for line, args in lines.items():
        if args is None:
            continue

        # Convert Origin/Destination into np.arrays
        origin = np.array([line.origin[0], line.origin[1], 0, 1])
        destination = np.array([line.destination[0], line.destination[1], 0, 1])

        # update the points based on the arg transformer
        adjusted = vec.resolve_vec_transformed_points(
            points=[origin, destination],
            vec_transformer_getter=vec_transformer_getter
        )

        # Draw the lines
        if args.color:
            w = args.width if args.width is not None else 1
            pygame.draw.line(surface,
                             args.color.value,
                             (adjusted[0][0], adjusted[0][1]),
                             (adjusted[1][0], adjusted[1][1]),
                             width=w)

        # Draw the points
        if args.control_point_args:
            putils.draw_points(
                points={
                    (x[0], x[1]): args.control_point_args for x in adjusted
                },
                surface=surface,
                vec_transformer_getter=vec_transformer_getter
            )

        # Draw Directionality
        if args.directionality_draw_args:
            help.draw_arrow(surface=surface,
                            color=args.directionality_draw_args.color,
                            start=(adjusted[0][0], adjusted[0][1]),
                            end=(adjusted[1][0], adjusted[1][1]),
                            arrow_height=args.directionality_draw_args.height,
                            arrow_width=args.directionality_draw_args.width)

        # Draw Label
        if args.draw_label_args is not None and args.label_text is not None:
            mid = ((adjusted[0][0] + adjusted[1][0]) / 2, (adjusted[0][1] + adjusted[1][1]) / 2)
            lutils.draw_label(hud=surface,
                              args=args.draw_label_args,
                              pos=mid,
                              text=args.label_text)


from typing import List
from coopstructs.vectors import Vector2, IVector
from coopstructs.geometry import Rectangle
import numpy as np

def calculate_fps(frame_times: List, ticks_last_frame: int):
    if len(frame_times) > 20:
        frame_times.pop(0)

    frame_times.append(ticks_last_frame)

    avg_sec_per_frame = sum(frame_times) / len(frame_times) / 1000.0
    fps = 1 / avg_sec_per_frame if avg_sec_per_frame > 0 else 0
    return fps

def game_area_coords_from_parent_coords(parent_coords: Vector2, game_area_surface_rectangle: Rectangle) -> Vector2:
    """Converts Global Coords into coords on the game area"""
    return Vector2(parent_coords.x - game_area_surface_rectangle.x, parent_coords.y - game_area_surface_rectangle.y)

def scaled_points_of_a_rect(rect, grid_pos: Vector2, draw_scale_matrix=None, margin: int = 1):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    ''' get the rectangle object representing the grid position that was input'''
    rect = Rectangle(x=(margin + rect.width) * grid_pos.x + margin
                     , y=(margin + rect.height) * grid_pos.y + margin
                     , height=rect.height
                     , width=rect.width)

    '''Convert the rectangle to a list of points at the 4 corners'''
    points = [(x[0], x[1], 0, 1) for x in rect.points_tuple()]

    '''Multiply the points by the transform matrix for drawing'''
    transformed_points = draw_scale_matrix.dot(np.transpose(points))  # Transpose the points to appropriately mutiply

    '''return the x and y position for all points on the rectangle'''
    return np.transpose(transformed_points)[:,
           :3]  # Re-Transpose the points back to remain in a "list of points" format


def scaled_points_of_points(points: List[Vector2], draw_scale_matrix=None):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    if len(points) == 0:
        return np.array([])
    '''Convert the point to a 4-dim point for multiplication'''
    normal_points = [(point.x, point.y, 0, 1) for point in points]

    '''Multiply the points by the transform matrix for drawing'''
    transformed_points = draw_scale_matrix.dot(
        np.transpose(normal_points))  # Transpose the points to appropriately mutiply

    return np.transpose(transformed_points)[:,
           :3]  # Re-Transpose the points back to remain in a "list of points" format


def viewport_point_on_plane(viewport_point: Vector2, game_area_rect, draw_scale_matrix=None, margin: int = 1):
    points_on_plane = scaled_points_of_a_rect(game_area_rect, Vector2(0, 0), draw_scale_matrix, margin=margin)
    point0 = points_on_plane[0]
    point1 = points_on_plane[1]
    point2 = points_on_plane[2]

    vec1 = point0 - point1
    vec2 = point2 - point1

    normal = np.cross(vec1, vec2)
    a = normal[0]
    b = normal[1]
    c = normal[2]
    d = a * point0[0] + b * point0[1] + c * point0[2]

    z_val = (d - a * viewport_point.x - b * viewport_point.y) / c

    return [viewport_point.x, viewport_point.y, z_val]


def scaled_points_to_normal_points(points: List[IVector], draw_scale_matrix=None):
    translated_points = [(point.x, point.y, 0, 1) for point in points]
    normal_array = scaled_array_to_normal_array(scaled_array=translated_points, draw_scale_matrix=draw_scale_matrix)
    normal_points = [Vector2(point[0], point[1]) for point in normal_array]

    return normal_points


def scaled_array_to_normal_array(scaled_array, draw_scale_matrix=None):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    draw_scale_matrix_inv = np.linalg.inv(draw_scale_matrix)
    normal_points = draw_scale_matrix_inv.dot(np.transpose(scaled_array))

    normal_points = np.reshape(normal_points, (-1, 2))

    return normal_points


def normal_points_to_scaled_points(points: List[Vector2], draw_scale_matrix=None):
    return scaled_points_of_points(points, draw_scale_matrix)


def mouse_in_plane_point(game_area_rect: Rectangle, draw_scale_matrix=None):
    # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
    """Gets the mouse position and converts it to a grid position"""
    mouse_game_area_coord = game_area_coords_from_parent_coords(parent_coords=mouse_pos_as_vector(),
                                                                game_area_surface_rectangle=game_area_rect)
    mouse_plane_point = viewport_point_on_plane(mouse_game_area_coord, game_area_rect, draw_scale_matrix, margin=1)

    return mouse_plane_point


def normal_mouse_position(game_area_rect: Rectangle, draw_scale_matrix=None) -> Vector2:
    # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
    """Gets the mouse position and converts it to a grid position"""
    # mouse_game_area_coord = game_area_coords_from_parent_coords(parent_coords=mouse_pos_as_vector(), game_area_surface_rectangle=game_area_rect)
    mouse_plane_point = mouse_in_plane_point(game_area_rect, draw_scale_matrix)
    points = np.array([mouse_plane_point[0], mouse_plane_point[1], mouse_plane_point[2], 1])

    normal_array = scaled_array_to_normal_array(points, draw_scale_matrix)

    return Vector2(normal_array[0][0], normal_array[0][1])


from typing import List

from coopgame.gameTemplate import GameTemplate, BuiltInSurfaceType
from coopgame.spriteHandling.spritePoseHandler import SpritePoseHandler
from coopgame.spriteHandling.spriteFolder import SpriteFolderHandler
from coopgame.spriteHandling.sprites import AnimatedPoseSprite
from cooptools.toggles import IntegerRangeToggleable
from coopstructs.vectors import Vector2
import pygame
import coopgame.pygamehelpers as help

SPRITE_DIRS = [
    r'C:\Users\tburns\Downloads\ninjaadventurenew\png',
    r'C:\Users\tburns\Downloads\freeknight\png',
    r'C:\Users\tburns\Downloads\zombiefiles\png\male',
    r'C:\Users\tburns\Downloads\zombiefiles\png\female',
    r'C:\Users\tburns\Downloads\craftpix-net-786503-free-spaceship-pixel-art-sprite-sheets\Fighter'
]

class SpriteTester(GameTemplate):

    def __init__(self):
        super().__init__()
        self.sprite_handler = SpritePoseHandler.from_spritefolder(
            SpriteFolderHandler(SPRITE_DIRS[4]))

        self.animation_cycle = IntegerRangeToggleable(min=10, max=500, step_size=25)

        self.anim_sprite = AnimatedPoseSprite(
            id='main',
            init_pos=Vector2(0, 0),
            pose_handler=self.sprite_handler,
            animation_cycle_ms=lambda: self.animation_cycle.value,
            width=int(self.screen_width / 3),
            height=int(self.screen_height)
        )

    def initialize_game(self):
        pass

    def draw(self, frames: int, debug_mode: bool = False):
        self.anim_sprite.blit(self.screen)
        current_pose_data = self.anim_sprite.pose_handler.get_current()
        help.draw_text(f"Current Pose: {current_pose_data['pose']}, "
                       f"anim_idx: {current_pose_data['animation_idx']}, "
                       f"anim cycle: {self.animation_cycle.value}"
                       , self.screen)


    def model_updater(self, delta_time_ms: int):
        pass

    def sprite_updater(self, delta_time_ms: int):
        self.anim_sprite.animate(delta_time_ms)

    def on_resize(self):
        pass

    def _increase_animation_cycle(self):
        self.animation_cycle.toggle(loop=False)

    def _decrease_animation_cycle(self):
        self.animation_cycle.toggle(reverse=True, loop=False)

    def register_keys(self):
        self.register_action_to_keys((pygame.K_LEFT,),
                                     lambda key_pressed_args: self.anim_sprite.pose_handler.decrement_pose(),
                                     react_while_holding=False)
        self.register_action_to_keys((pygame.K_RIGHT,),
                                     lambda key_pressed_args: self.anim_sprite.pose_handler.increment_pose(),
                                     react_while_holding=False)
        self.register_action_to_keys((pygame.K_UP,),
                                     lambda key_pressed_args: self._decrease_animation_cycle(),
                                     react_while_holding=False)
        self.register_action_to_keys((pygame.K_DOWN,),
                                     lambda key_pressed_args: self._increase_animation_cycle(),
                                     react_while_holding=False)
    def update_built_in_surfaces(self, surface_types: List[BuiltInSurfaceType]):
        pass

if __name__ == "__main__":
    game = SpriteTester()
    game.main()


import pygame
from typing import List, Tuple, Callable, Any
from coopgame.particleSystem.particleGenerationArgs import PhaseArgs
from dataclasses import dataclass
from coopgame.particleSystem.protocol import UpdateRenderableProtocol
from coopgame.particleSystem.callables import UpdateRenderableProtocol_provider

class EmitterPhaseController:
    def __init__(self,
                 args: PhaseArgs,
                 context = None
                 ):
        self.args = args
        self.context = context


        self.phases = self.resolved_phases()
        self.phase_idx = None

    def resolved_phases(self):
        phases = []
        for idx, x in enumerate(self.args.phases):
            if idx==0:
                phases.append((x.resolve(None), x))
            else:
                phases.append((x.resolve(phases[-1][0]), x))

        return phases

    def update(self, delta_ms):
        epoch_start_phase = self.phase_idx

        if self.phase_idx is None:
            self.phase_idx = 0

        if self.phase_idx >= len(self.phases):
            return

        emitter, phase_args = self.phases[self.phase_idx]

        emitter.update(delta_ms)

        if phase_args.phase_end_trigger is not None and phase_args.phase_end_trigger(self.phases[self.phase_idx][0]):
            self.phase_idx += 1

        if self.phase_idx >= len(self.phases) and self.args.loop:
            self.phase_idx = 0

        #handle callbacks
        if epoch_start_phase is None:
            # play first start callback
            self.phases[self.phase_idx][1].phase_start_callback(self.phases[self.phase_idx][0])
        elif self.phase_idx != epoch_start_phase:
            # end last callback
            self.phases[epoch_start_phase][1].phase_end_callback(self.phases[epoch_start_phase][0])
            # start new callback
            if self.phase_idx < len(self.phases):
                self.phases[self.phase_idx][1].phase_start_callback(self.phases[self.phase_idx][0])


    def render(self, surf: pygame.Surface):
        if self.phase_idx < len(self.phases):
            emitter, _ = self.phases[self.phase_idx]
            emitter.render(surf)

    def n_particles(self):
        if self.phase_idx < len(self.phases):
            return self.phases[self.phase_idx][0].n_particles()
        else:
            return 0

    @property
    def N_Particles(self) -> int:
        return self.n_particles()

    @property
    def Expired(self) -> bool:
        return self.phase_idx > len(self.phases)


import pygame
from coopstructs.geometry import Rectangle
from coopgame.colors import Color
from coopstructs.geometry.vectors.vectorN import Vector2, VectorN
import numpy as np
from typing import List, Dict, Tuple
import logging
from coopgame.label_drawing.pyLabel import TextAlignmentType
import cooptools.geometry_utils.vector_utils as vec
import cooptools.geometry.circles.utils as circ
import math
from cooptools.anchor import Anchor2D, CardinalPosition
from cooptools.matrixManipulation import point_transform_3d
from coopgame.pointdrawing.point_draw_utils import draw_points, DrawPointArgs
import cooptools.matrixManipulation as mm

def mouse_pos() -> vec.FloatVec:
    return pygame.mouse.get_pos()

def mouse_pos_as_vector() -> Vector2:
    """ Get the global coords of the mouse position and convert them to a Vector2 object"""
    return Vector2(*mouse_pos())

def draw_box(surface: pygame.Surface,
             rect: Rectangle,
             color: Color = None,
             width: int = 0,
             outline_color: Color = None,
             anchor_draw_args: DrawPointArgs=None,
             corner_draw_args: DrawPointArgs=None,
             vec_transform_getter: vec.VecTransformerGetter = None
             ):

    if color:
        pygame.draw.rect(surface, color.value, (rect.TopLeft[0], rect.TopLeft[1], rect.width, rect.height), width)
    if outline_color:
        pygame.draw.rect(surface, outline_color.value, (rect.TopLeft[0], rect.TopLeft[1], rect.width, rect.height), 1)
    if anchor_draw_args:
        draw_points(
            surface=surface,
            points={rect.anchor.pos(): anchor_draw_args},
            vec_transformer_getter=vec_transform_getter
        )
    if corner_draw_args:
        draw_points(
            surface=surface,
            points={x: corner_draw_args for x in rect.corner_generator()},
            vec_transformer_getter=vec_transform_getter
        )

def draw_polygon(surface: pygame.Surface,
                 points,
                 color: Color,
                 width: int = 0,
                 alpha: int = 255,
                 vec_transform_getter: vec.VecTransformerGetter = None):
    if type(points[0]) in [VectorN, Vector2]:
        points = [point.as_tuple() for point in points]

    scaled = vec.resolve_vec_transformed_points(
        points=points,
        vec_transformer_getter=vec_transform_getter
    )

    # TODO: this is slower, but may support alpha. It needs to be evaluated further
    # minx = min(point[0] for point in points)
    # maxx = max(point[0] for point in points)
    # miny = min(point[1] for point in points)
    # maxy = max(point[1] for point in points)
    #
    # points = [(point[0] - minx, point[1] - miny) for point in points]
    # s = pygame.Surface((int(maxx - minx), int(maxy - miny)), pygame.SRCALPHA)
    # c_with_alpha = (color.value[0], color.value[1], color.value[2], alpha)
    # pygame.draw.polygon(s, c_with_alpha, points, width)
    # surface.blit(s, (minx, miny))

    c_with_alpha = (color.value[0], color.value[1], color.value[2], alpha)
    points = [x[:2] for x in scaled]
    pygame.draw.polygon(surface, c_with_alpha, points, width)

def draw_arrow(surface: pygame.Surface,
               start: vec.FloatVec,
               end: vec.FloatVec,
               color: Color,
               arrow_height: int = 5,
               leader_line_width: int = 1,
               leader_line: bool = False,
               arrow_width: int = 5):
    if leader_line:
        pygame.draw.line(surface, color.value, start, end, width=leader_line_width)

    if start == end:
        return

    arrow_points = [end, (end[0] - arrow_width / 2, end[1] - arrow_height), (end[0] + arrow_width / 2, (end[1] - arrow_height))]

    unit_backwards = vec.unit_vector(vec.vector_between(end, start))
    unit_backwards = (unit_backwards[0], -unit_backwards[1])  # negative to account for the pygame inverse Y-orientation
    unit_angle = circ.rads_between(unit_backwards)
    adjustment = - math.pi / 2
    rotation_angle = -(unit_angle + adjustment)  # negative to account for the pygame inverse Y-orientation

    rotated_points = []
    for point in arrow_points:
        rotated_points.append(circ.rotated_point(point, (end), rotation_angle))

    pygame.draw.polygon(surface, color.value, rotated_points)

def game_area_coords_from_parent_coords(parent_coords: Vector2, game_area_surface_rectangle: Rectangle) -> Vector2:
    """Converts Global Coords into coords on the game area"""
    return Vector2(parent_coords.x - game_area_surface_rectangle.x, parent_coords.y - game_area_surface_rectangle.y)


def scaled_points_of_a_rect(rect,
                            vec_transformer_getter: vec.VecTransformerGetter) -> List[Vector2]:
    ''' get the rectangle object representing the grid position that was input'''
    return scaled_points(rect.Corners,
                         vec_transformer_getter=vec_transformer_getter)

def scaled_points(points: List[Vector2],
                  vec_transformer_getter: vec.VecTransformerGetter=None) -> List[Vector2]:

    if vec_transformer_getter is None:
        return points

    points = vec.resolve_vec_transformed_points(
        points=[x.as_tuple() for x in points],
        vec_transformer_getter=vec_transformer_getter
    )

    return [Vector2.from_tuple(point) for point in points]


def viewport_point_on_plane(viewport_point: Vector2,
                            game_area_rect,
                            vec_transformer_getter:vec.VecTransformerGetter=None) -> vec.FloatVec:
    points_on_plane = scaled_points_of_a_rect(game_area_rect,
                                              vec_transformer_getter=vec_transformer_getter)

    projected = vec.points_projected_to_plane(
        [viewport_point.as_tuple()],
        plane_points=points_on_plane
    )[0]

    return projected

def scaled_points_to_normal_points(points: List[Vector2], draw_scale_matrix=None):
    translated_points = [(point.x, point.y, 0, 1) for point in points]
    normal_array = scaled_array_to_normal_array(scaled_array=translated_points, draw_scale_matrix=draw_scale_matrix)
    normal_points = [Vector2(point[0], point[1]) for point in normal_array]

    return normal_points


def scaled_array_to_normal_array(scaled_array, draw_scale_matrix=None):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    draw_scale_matrix_inv = np.linalg.inv(draw_scale_matrix)
    normal_points = draw_scale_matrix_inv.dot(np.transpose(scaled_array))

    normal_points = np.reshape(normal_points, (-1, 2))

    return normal_points


def normal_points_to_scaled_points(points: List[Vector2], draw_scale_matrix=None):
    return scaled_points(points, draw_scale_matrix)


def mouse_in_plane_point(game_area_rect: Rectangle, draw_scale_matrix=None):
    # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
    """Gets the mouse position and converts it to a grid position"""
    mouse_game_area_coord = game_area_coords_from_parent_coords(parent_coords=mouse_pos_as_vector(),
                                                                game_area_surface_rectangle=game_area_rect)
    mouse_plane_point = viewport_point_on_plane(mouse_game_area_coord, game_area_rect, draw_scale_matrix)

    return mouse_plane_point


def normal_mouse_position(game_area_rect: Rectangle, draw_scale_matrix=None) -> Vector2:
    # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
    """Gets the mouse position and converts it to a grid position"""
    # mouse_game_area_coord = game_area_coords_from_parent_coords(parent_coords=mouse_pos_as_vector(), game_area_surface_rectangle=game_area_rect)
    mouse_plane_point = mouse_in_plane_point(game_area_rect, draw_scale_matrix)
    points = np.array([mouse_plane_point[0], mouse_plane_point[1], mouse_plane_point[2], 1])

    normal_array = scaled_array_to_normal_array(points, draw_scale_matrix)

    return Vector2(normal_array[0][0], normal_array[0][1])

def calculate_fps(frame_times: List):
    avg_sec_per_frame = sum(frame_times) / len(frame_times) / 1000.0
    fps = 1 / avg_sec_per_frame if avg_sec_per_frame > 0 else 0
    return fps

def init_surface(dims):
    surface = pygame.Surface(dims).convert()
    surface.set_colorkey(Color.BLACK.value)
    return surface

def draw_mouse_pos(args: DrawPointArgs,
                   screen: pygame.Surface):
    draw_points(
        {tuple(int(x) for x in mouse_pos()): args},
        surface=screen,
    )



import logging
import threading

import numpy as np
import pygame
from coopstructs.geometry import Rectangle
from coopgame.colors import Color
from coopgame.pygbutton import PygButton
from cooptools.decor import timer, try_handler

from typing import Callable, List, Dict, Optional, Any
import coopgame.pygamehelpers as help
from coopgame.spriteHandling.sprites import AnimatedSprite
from coopstructs.geometry import Rectangle as cRect
from coopgame.monitoredClassLogger import MonitoredClassLogger
from cooptools.coopEnum import CoopEnum
from enum import auto
import inspect
from coopgame.pygame_k_constant_names import pygame_key_mapper, PyKeys, PyMouse
from coopgame.gameTimeTracker import GameTimeTracker
from cooptools.toggles import BooleanToggleable
from cooptools.timedDecay import Timer
from coopgame.label_drawing.pyLabel import TextAlignmentType
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputEvent, InputEventType, CallbackPackage
from coopgame.logger import logger
from cooptools.decor import TimeableProtocol
import coopgame.label_drawing.label_drawing_utils as labutils
from cooptools.coopEnum import CardinalPosition

DEBUG_COLORKEY = Color.DARK_SLATE_GRAY
DEBUGGER_MASK_ALPHA = 65
key_mapper = Callable[[InputState], Any]
DEBGUG_LABEL_ARGS = labutils.DrawLabelArgs(
    color=Color.LEMON_CHIFFON
)

class BuiltInSurfaceType(CoopEnum):
    BACKGROUND = auto()
    FOREGROUND = auto()
    HELP = auto()
    DEBUGGER = auto()
    DEBUGGER_MASK = auto()

class TimeTracker(TimeableProtocol):
    def __init__(self):
        self.internally_tracked_times = {}

tt = TimeTracker()

TransformMatrixProvider: Callable[[], np.array]


class GameTemplate:

    def __init__(self,
                 fullscreen: bool = False,
                 screen_width: int = 1000,
                 screen_height: int = 500,
                 max_fps: int = 120,
                 debug_mode=False,
                 input_state_handler: InputStateHandler = None,
                 log_stats_interval_ms: int = 5000):
        self.init_screen_width = screen_width
        self.init_screen_height = screen_height
        self.log_stats_interval_ms = log_stats_interval_ms

        self._lock = threading.RLock()

        self.fullscreen = fullscreen
        self.screen: pygame.Surface = None
        self._init_screen(self.fullscreen, self.screen_width, self.screen_height)
        self._monitored_class_logger = MonitoredClassLogger()

        self.debug_mode_toggle = BooleanToggleable(default=debug_mode)


        self.input_state_handler = input_state_handler if input_state_handler is not None else \
            InputStateHandler(quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
                              debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
                              fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()))

        self.game_time_tracker = GameTimeTracker(max_fps=max_fps)

        self.running = False

        self.buttons = {}
        self.sprites = {}

        # build in surfaces
        self.built_in_surfaces: Dict[BuiltInSurfaceType, Optional[pygame.Surface]] = {
            BuiltInSurfaceType.BACKGROUND: None,
            BuiltInSurfaceType.FOREGROUND: None,
            BuiltInSurfaceType.HELP: None,
            BuiltInSurfaceType.DEBUGGER: None
        }

        self.game_time_tracker.set_start()
        self.start_timers()
        # self.log_stats_refresher = DataRefresher('log_stats',
        #                                          refresh_callback=lambda: logger.info(
        #                                              f"FPS: {int(self.game_time_tracker.fps) if self.game_time_tracker.fps else None}"),
        #                                          refresh_interval_ms=1000)
        pygame.init()

    def start_timers(self):
        self.debug_surface_update_timer = Timer(100,
                                                start_on_init=True,
                                                as_async=True)
        self.log_stats_timer = Timer(5000,
                                     start_on_init=True,
                                     as_async=True)

    def quit(self):
        self.running = False

    def toggle_fullscreen(self):
        self.fullscreen = not self.fullscreen
        self._init_screen(self.fullscreen, self.screen_width, self.screen_height)

    def _init_screen(self, fullscreen=None, screen_width=None, screen_height=None):
        if fullscreen:
            self.screen = pygame.display.set_mode((0, 0), pygame.FULLSCREEN)
        elif screen_width and screen_height:
            self.screen = pygame.display.set_mode((self.init_screen_width, self.init_screen_height))
        else:
            self.screen = pygame.display.set_mode((0, 0))

    def register_button(self, id, text, callback, postion_rect):
        self.buttons[id] = PygButton(postion_rect, caption=text, callback=callback)

    @timer(logger=logger, time_tracking_class=tt)
    def main(self, debug: bool = False):
        self.debug_mode_toggle.set_value(debug)

        self.init_builtin_surfaces()
        self.initialize_game()

        self.running = True

        while self.running:
            self._update()
            self._draw(frames=self.game_time_tracker.frames)

        pygame.quit()

    def _check_log_stats(self):
        if self.log_stats_timer.Finished:
            logger.info(f"FPS: {int(self.game_time_tracker.fps) if self.game_time_tracker.fps else None}")
            self.log_stats_timer.reset(start=True)

    @timer(logger=logger, time_tracking_class=tt)
    def _update(self):
        """:return
            Update environment based on time delta and any input
        """

        ''' Calculate the ticks between update calls so that the update functions can handle correct time deltas '''
        delta_time_ms = self.game_time_tracker.update()

        # '''Log Stats'''
        # self._monitored_class_logger.check_and_log(logger=logger, delta_time_ms=delta_time_ms, log_interval_ms=self.log_stats_interval_ms)
        # self.log_stats_refresher.check_and_refresh()

        '''Handle Events'''
        """ Handling events should come before updating the model so that any updates to input, etc
        are appropriately handled in the model prior to drawing. O/w, the draw can be written explicitly
        on the last state (not model) and can cause it to be out of order"""
        self._handle_events(delta_time_ms)

        '''Update Model'''
        self._model_updater(delta_time_ms)

        '''Update Sprites'''
        self.sprite_updater(delta_time_ms)

        '''Animate Sprites'''
        self._sprite_animator(delta_time_ms)

    @timer(logger=logger, time_tracking_class=tt)
    def _handle_events(self, delta_time_ms: int):
        """:return
            handle all of the registered events that have been captured since last iteration
        """

        '''Mouse Pos'''
        inp_state = InputState(delta_ms=delta_time_ms)
        inp_state.register(mouse_pos=help.mouse_pos_as_vector().as_tuple())

        '''Get next event'''
        for event in pygame.event.get():
            '''Check and handle button press'''
            self.handle_buttons(event)

            '''Debug Printer'''
            if event.type not in (0, 1, 4, 6):
                logger.debug(f"Pygame EventType: {event.type}")

            '''Event Type Switch'''
            if event.type == pygame.QUIT:
                self.quit()
            elif event.type == pygame.KEYDOWN:
                inp_state.register(events=[InputEvent(event_type=InputEventType.DOWN, input_key=PyKeys.by_val(event.key))])
            elif event.type == pygame.MOUSEBUTTONDOWN:
                inp_state.register(events=[InputEvent(event_type=InputEventType.DOWN, input_key=PyMouse.by_val(event.button))])
            elif event.type == pygame.KEYUP:
                inp_state.register(events=[InputEvent(event_type=InputEventType.UP, input_key=PyKeys.by_val(event.key))])
            elif event.type == pygame.MOUSEBUTTONUP:
                inp_state.register(events=[InputEvent(event_type=InputEventType.UP, input_key=PyMouse.by_val(event.button))])
            elif event.type == pygame.WINDOWSIZECHANGED:
                self._on_resize()
            else:
                logger.debug(f"Unhandled event: {event.type}")

        '''Handle Input State'''
        self.input_state_handler.handle_input(inp_state)

    def register_action_to_keys(self, keys_tuple, func: key_mapper, react_while_holding: bool = False):
        """
            Takes a tuple of keys (integers) and maps that key combo to a callable function

            :param keys_tuple a list of keys (integers) that will be mapped to the input callable. Note that a single
            value Tuple is input as ([key],) *note the comma
            :param func a callable that is mapped to the key combo
        """
        self._key_handlers[keys_tuple] = (func, react_while_holding)

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def handle_buttons(self, event):
        for id, button in self.buttons:
            if 'click' in button.handleEvent(event):
                button.callback()

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def _model_updater(self, delta_time_ms: int):
        return self.model_updater(delta_time_ms)

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def initialize_game(self):
        raise NotImplementedError()

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def handle_hover_over(self, input_state: InputState):
        raise NotImplementedError()

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def draw(self, frames: int, debug_mode: bool = False):
        raise NotImplementedError()

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def model_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def sprite_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    def _on_resize(self):
        logger.info(f"Window Resized [{self.screen.get_width()}x{self.screen.get_height()}]")
        self.init_builtin_surfaces()
        self.on_resize()

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def on_resize(self):
        raise NotImplementedError()

    @timer(logger=logger, time_tracking_class=tt)
    @try_handler(logger=logger)
    def update_built_in_surfaces(self, surface_types: List[BuiltInSurfaceType]):
        raise NotImplementedError()

    @timer(logger=logger, time_tracking_class=tt)
    def _draw(self, frames: int):
        self.screen.fill(Color.BLACK.value)
        self.draw(frames, self.debug_mode_toggle.value)

        '''Draw Sprites'''
        self.sprite_drawer(self.screen)

        if self.debug_mode_toggle.value:
            self._check_update_debugger_surface()
            self.screen.blit(self.built_in_surfaces[BuiltInSurfaceType.DEBUGGER_MASK], (0, 0))
            self.screen.blit(self.built_in_surfaces[BuiltInSurfaceType.DEBUGGER], (0, 0))

        # Update the display
        pygame.display.flip()

    def _check_update_debugger_surface(self):
        if self.debug_surface_update_timer.Finished:
            self._update_debugger_surface()
            self.debug_surface_update_timer.reset(start=True)

    def _update_debugger_surface(self):

        deb_surf = self.built_in_surfaces[BuiltInSurfaceType.DEBUGGER]
        if deb_surf is None:
            return

        deb_surf.fill(DEBUG_COLORKEY.value)
        deb_deb = False

        # draw sprites
        labutils.draw_label(
            hud=deb_surf,
            text=f"Sprite Count: {len(self.sprites)}",
            args=DEBGUG_LABEL_ARGS.with_(
                alignment=TextAlignmentType.TOPRIGHT
            ),
            offset_rect=cRect.from_tuple((self.screen.get_width(),
                                          self.screen.get_height() - 40,
                                          400,
                                          20),
                                         anchor_cardinality=CardinalPosition.BOTTOM_RIGHT,
                                         inverted_y=True),
            deb=deb_deb
        )

        #draw fps
        labutils.draw_label(
            hud=deb_surf,
            text=f"FPS: {str(int(self.game_time_tracker.fps))}",
            args=DEBGUG_LABEL_ARGS.with_(
                alignment=TextAlignmentType.TOPRIGHT
            ),
            offset_rect=cRect.from_tuple((self.screen.get_width(),
                                          self.screen.get_height() - 60,
                                          400,
                                          20),
                                         anchor_cardinality=CardinalPosition.BOTTOM_RIGHT,
                                         inverted_y=True),
            deb=deb_deb
        )

        # draw mouse coords
        labutils.draw_label(
            hud=deb_surf,
            text=f"M:<{help.mouse_pos()}>",
            args=DEBGUG_LABEL_ARGS.with_(
                alignment=TextAlignmentType.TOPRIGHT
            ),
            offset_rect=cRect.from_tuple((self.screen.get_width(),
                                          self.screen.get_height(),
                                          400,
                                          20),
                                         anchor_cardinality=CardinalPosition.BOTTOM_RIGHT,
                                         inverted_y=True),
            deb=deb_deb
        )

        # draw game time
        game_time_s = self.game_time_tracker.ticks / 1000
        hrs = int(game_time_s / 3600)
        min = int((game_time_s - hrs * 3600) / 60)
        s = round(game_time_s - hrs * 3600 - min * 60, 2)
        txt = f"GameTime: {hrs} hrs, {min} min, {s} sec"
        labutils.draw_label(
            hud=deb_surf,
            text=txt,
            args=DEBGUG_LABEL_ARGS.with_(
                alignment=TextAlignmentType.TOPRIGHT
            ),
            offset_rect=cRect.from_tuple((self.screen.get_width(),
                                          self.screen.get_height() - 20,
                                          400,
                                          20),
                                         anchor_cardinality=CardinalPosition.BOTTOM_RIGHT,
                                         inverted_y=True),
            deb=deb_deb
        )

        self.draw_monitoredclass_stats(surface=deb_surf, deb=deb_deb)

    def draw_monitoredclass_stats(self,
                                  surface: pygame.Surface,
                                  deb:bool=False):
        offset = 0
        offset = labutils.draw_dict(dict_to_draw=tt.internally_tracked_times,
                                    surface=surface,
                                    total_game_time_sec=self.game_time_tracker.ticks / 1000,
                                    title=f"RunTime Stats",
                                    draw_args=DEBGUG_LABEL_ARGS.with_(
                                        alignment=TextAlignmentType.TOPLEFT
                                    ),
                                    deb=deb
        )

    def _sprite_animator(self,
                         delta_time_ms: int):

        for name, sprite in self.sprites.items():
            if type(sprite) == AnimatedSprite:
                sprite.animate(delta_time_ms)

    @timer(logger=logger, time_tracking_class=tt)
    def sprite_drawer(self,
                      surface: pygame.surface):

        sprites = list(self.sprites.values())
        sprites.sort(key=lambda x: x.bottom_center_pos.y)
        for sprite in sprites:
            sprite.blit(surface, display_handle=self.debug_mode_toggle.value, display_rect=self.debug_mode_toggle.value)

    # def register_monitored_classes(self, new_classes: List[TimeableProtocol]):
    #     self._monitored_class_logger.register_classes(new_classes)

    def init_builtin_surfaces(self, types: List[BuiltInSurfaceType] = None, colorkey: Color = None):
        if colorkey is None: colorkey = DEBUG_COLORKEY
        if types is None: types = [e for e in BuiltInSurfaceType]

        for type in types:
            self.built_in_surfaces[type] = pygame.Surface(self.screen.get_size()).convert()
            self.built_in_surfaces[type].set_colorkey(colorkey.value)
            if type == BuiltInSurfaceType.DEBUGGER_MASK:
                self.built_in_surfaces[type].set_alpha(DEBUGGER_MASK_ALPHA)

    @timer(logger=logger, time_tracking_class=tt)
    def _update_help_surface(self):
        text = self._help_txt()

        self.built_in_surfaces[BuiltInSurfaceType.HELP].fill(Color.WHEAT.value)

        font = pygame.font.Font(None, 20)
        fontsize = font.get_height()
        offSet = 0

        for idx, line in enumerate(text):
            help.draw_text(line,
                           self.built_in_surfaces[BuiltInSurfaceType.HELP],
                           font=font,
                           offset_rect=Rectangle.from_tuple((0,
                                                 idx * fontsize + offSet,
                                                 fontsize + 3,
                                                 self.built_in_surfaces[BuiltInSurfaceType.HELP].get_width())))

    def _help_txt(self):

        txt = []
        txt.append("HELP")
        txt.append("\n--------")
        txt.append("\n")
        for k, v in self._key_handlers.items():
            if len(v) > 2:
                callback_txt = v[2]
            elif v[0].__name__ == "<lambda>":
                callback_txt = inspect.getsource(v[0])
                callback_txt = callback_txt.split("lambda:", 1)[1].split('(', 1)[0].replace("self.", "")
            else:
                callback_txt = v[0].__name__
            k_tup = [pygame_key_mapper(x) for x in k]
            txt.append(f"{k_tup} -- {callback_txt}")

        return txt

    @property
    def screen_width(self):
        return self.screen.get_width() if self.screen else self.init_screen_width

    @property
    def screen_height(self):
        return self.screen.get_height() if self.screen else self.init_screen_height

    @property
    def game_area_rectangle(self):
        return Rectangle.from_tuple((0, 0, self.screen.get_width(), self.screen.get_height()))

    @property
    def ScreenCenter(self):
        return self.game_area_rectangle.Center

if __name__ == "__main__":
    import coopgame.pointdrawing.point_draw_utils as pdu

    class HelpTest(GameTemplate):

        def __init__(self):
            super().__init__()

        def draw(self, frames: int, debug_mode: bool = False):
            pdu.draw_points(
                points={
                    (self.ScreenCenter): pdu.DrawPointArgs(
                        color=Color.BLUE,
                        outline_color=Color.ORANGE,
                        radius=50,
                        outline_width=4
                    )
                },
                surface=self.screen
            )
            help.draw_arrow(self.screen,
                            self.ScreenCenter,
                            help.mouse_pos_as_vector().as_tuple(),
                            Color.GREEN)

    logging.basicConfig(level=logging.INFO)
    g = HelpTest()
    g.main()


from cooptools.colors import Color
from typing import Protocol, Tuple, Union, List
from coopgame.particleSystem.callables import color_callable
import coopgame.particleSystem.utils as utils


def _interpolate_int(start, end, progress):
    return int(start + (end - start) * progress)

def _resolve_color(color: Union[Color, color_callable]):
    if callable(color):
        return color()
    else:
        return color

def interpolate_at_t(t: int,
                     start_color: Union[Color, color_callable],
                     end_color: Union[Color, color_callable],
                     lifespan_ms: int) -> Tuple[int, int, int]:
    progress = min(t / lifespan_ms, 1)

    start = _resolve_color(start_color)
    end = _resolve_color(end_color)

    r = _interpolate_int(start.value[0], end.value[0], progress)
    g = _interpolate_int(start.value[1], end.value[1], progress)
    b = _interpolate_int(start.value[2], end.value[2], progress)

    return r, g, b

class ColorBehaviorProtocol(Protocol):
    def resolve_at_t(self, t: int) -> Tuple[int, int, int]:
        pass


class PulseColorBehavior:
    def __init__(self,
                 colors: List[Color],
                 pulse_duration_ms: int,
                 ):
        self.colors = colors
        self.pulse_duration_ms = pulse_duration_ms

    def resolve_at_t(self, t: int) -> Tuple[int, int, int]:
        color_changes = t / self.pulse_duration_ms

        current_start_color_idx = int(color_changes) % len(self.colors)
        current_end_color_idx = current_start_color_idx + 1

        if current_end_color_idx >= len(self.colors):
            current_end_color_idx = 0

        return interpolate_at_t(int((color_changes - int(color_changes)) * 1000),
                                self.colors[current_start_color_idx],
                                self.colors[current_end_color_idx],
                                1000
        )

class InterpolateColorBehavior:
    def __init__(self, start_color: Union[Color, color_callable], end_color: Union[Color, color_callable], lifespan_ms: int):
        self.start = start_color
        self.end = end_color
        self.lifespan_ms = lifespan_ms

    def resolve_at_t(self, t: int) -> Tuple[int, int, int]:
        return interpolate_at_t(
            t=t,
            start_color=self.start,
            end_color=self.end,
            lifespan_ms=self.lifespan_ms
        )




import numpy as np

def rotation_unit_vector(axis: np.array):
    return axis / (axis ** 2).sum() ** 0.5

def translationMatrix(dx=0, dy=0, dz=0):
    """ Return matrix for translation along vector (dx, dy, dz). """
    return np.array([[1, 0, 0, dx],
                     [0, 1, 0, dy],
                     [0, 0, 1, dz],
                    [0, 0, 0, 1]])

def scaleMatrix(sx=0, sy=0, sz=0):
    """ Return matrix for scaling equally along all axes centred on the point (cx,cy,cz). """
    return np.array([[sx, 0, 0, 0],
                     [0, sy, 0, 0],
                     [0, 0, sz, 0],
                     [0, 0, 0, 1]])

def rotateXMatrix(radians):
    """ Return matrix for rotating about the x-axis by 'radians' radians """
    c = np.cos(radians)
    s = np.sin(radians)
    return np.array([[1, 0, 0, 0],
                     [0, c, s, 0],
                     [0, -s, c, 0],
                     [0, 0, 0, 1]])

def rotateYMatrix(radians):
    """ Return matrix for rotating about the y-axis by 'radians' radians """
    c = np.cos(radians)
    s = np.sin(radians)
    return np.array([[c, 0, -s, 0],
                     [0, 1, 0, 0],
                     [s, 0, c, 0],
                     [0, 0, 0, 1]])

def rotateZMatrix(radians):
    """ Return matrix for rotating about the z-axis by 'radians' radians """
    c = np.cos(radians)
    s = np.sin(radians)
    return np.array([[c, s, 0, 0],
                     [-s, c, 0, 0],
                     [0, 0, 1, 0],
                     [0, 0, 0, 1]])

def rotateAroundAxis(rotationPoint, rotationAxis, radians):
    # http://www.fastgraph.com/makegames/3drotation/
    # https://sites.google.com/site/glennmurray/Home/rotation-matrices-and-formulas/rotation-about-an-arbitrary-axis-in-3-dimensions
    c = np.cos(radians)
    s = np.sin(radians)
    t = 1 - np.cos(radians)

    u = rotationAxis[0]
    v = rotationAxis[1]
    w = rotationAxis[2]


    ''' Build the transposes into the matrix'''
    x = rotationPoint[0]
    y = rotationPoint[1]
    z = rotationPoint[2]
    calculated_matrix = np.array([[u**2 + (v**2 + w**2) * c, u*v*t-w*s,              u*w*t+v*s,              (x*(v**2 + w**2) - u*(y*v + z*w))*t + (y*w-z*v)*s],
                                 [u*v*t + w*s,              v**2 + (u**2 + w**2)*c, v*w*t-u*s,              (y*(u**2 + w**2) - v*(x*u + z*w))*t + (z*u-x*w)*s],
                                 [u*w*t - v*s,              v*w*t + u*s,            w**2 + (u**2 + v**2)*c, (z*(u**2 + v**2) - w*(x*u + y*v))*t + (x*v-y*u)*s],
                                 [0,                        0,                      0,                      1]])

    # r00 = t * u * u + c
    # r01 = t * u * v - s * w
    # r02 = t * u * w + s * u
    # r10 = t * u * v + s * w
    # r11 = t * v * v + c
    # r12 = t * v * w - s * u
    # r20 = t * u * w - s * u
    # r21 = t * v * w + s * u
    # r22 = t * w * w + c
    # r03 = (x*(v**2 + w**2) - u*(y*v + z*w))*t + (y*w-z*v)*s
    # r13 = (y*(u**2 + w**2) - v*(x*u + z*w))*t + (z*u-x*w)*s
    # r23 = (z*(u**2 + v**2) - w*(x*u + y*v))*t + (x*v-y*u)*s
    # calculated_matrix = np.array([[r00, r01, r02, r03],
    #                               [r10, r11, r12, r13],
    #                               [r20, r21, r22, r23],
    #                               [0,   0,   0,   1]])

    return calculated_matrix

def get_grid_rotation_matrix(rotation_axis_x, rotation_axis_z, x_rot_rads, z_rot_rads):

    # Rot Z
    orientation = np.array([0, 0, 1])
    ruv = rotation_unit_vector(orientation)
    matrixz = rotateAroundAxis((rotation_axis_x, rotation_axis_z, 0), ruv, z_rot_rads)

    # Rot X
    orientation = np.array([1, 0, 0])
    ruv = rotation_unit_vector(orientation)
    matrixx = rotateAroundAxis((rotation_axis_x, rotation_axis_z, 0), ruv,
                                  x_rot_rads)
    matrix = matrixx.dot(matrixz)
    return matrix



    # translationM = translationMatrix(-rotationPoint[0], -rotationPoint[1], -rotationPoint[2])
    # print (f"translationM: \n{translationM}")
    # # return translationM
    # translationMInv = translationMatrix(rotationPoint[0], rotationPoint[1], rotationPoint[2])
    #
    #
    # matrix =  np.array([[t * u ** 2 + c,    t * u * v - s * w,  t * u * w + s * v,  0],
    #                     [t * u * v + s * w, t * v**2 + c,       t*v*w-s*u,          0],
    #                     [t*u*w - s*v,       t*v*w+s*u,          t*w**2+c,           0],
    #                     [0,                 0,                  0,                  1]])
    #
    #
    # # return matrix
    # with np.printoptions(precision=3, suppress=True):
    #     print(f"Rotation \n{matrix}")
    #     print(matrix.dot(translationM))
    #     finalTransform = translationMInv.dot(matrix.dot(translationM))
    #     # finalTransform = translationM.dot(matrix).dot(translationMInv)
    #     print(f"Final Transform: \n{finalTransform}")
    #
    #     print(f"clc: {calculated_matrix}")
    #     print(f"mat: {finalTransform}")
    #
    # return finalTransform


def rotateAroundPoint( point, rotationVector):

    translationM = translationMatrix(-point[0], -point[1], -point[2])


    translationMInv = translationMatrix(point[0], point[1], point[2])
    rX = rotateXMatrix(rotationVector[0])
    rY = rotateYMatrix(rotationVector[1])
    rZ = rotateZMatrix(rotationVector[2])

    return translationM.dot(rX.dot(rY.dot(rZ.dot(translationMInv))))

def scaleAroundPoint(point, scalarVector):
    translationM = translationMatrix(-point[0], -point[1], -point[2])
    translationMInv = translationMatrix(point[0], point[1], point[2])
    scaleM = scaleMatrix(*scalarVector)
    matrix = translationMInv.dot(scaleM.dot(translationM))

    return matrix



if __name__ == "__main__":
    import math
    import time
    from coopstructs.vectors import Vector3
    rot_point = (1, 100, 0)
    start = (1, 2, 0, 1)
    rot_axis = Vector3(1, 1, 1).unit().as_tuple()
    rads = math.pi / 2

    R = rotateAroundAxis(rot_point, rot_axis, rads)
    # print(np.array(start))
    # current = (np.dot(R, np.transpose(start)).round(3))
    # print(current)
    # current = (np.dot(R, np.transpose(current)).round(3))
    # print(current)
    # current = (np.dot(R, np.transpose(current)).round(3))
    # print(current)
    # current = (np.dot(R, np.transpose(current)).round(3))
    # print(current)

    current = np.array(start)
    while True:
        current = (np.dot(R, np.transpose(current)).round(3))
        print(current.round(1))
        time.sleep(.5)


import coopgame.renderedObjectHandling.MatrixManipulation as mm
from coopgame.renderedObjectHandling.objectOrientationArgs import ObjectOrientationArgs
import numpy as np



def object_orientation_matrix(orientation_args: ObjectOrientationArgs, invert:bool = False):
    """
    In order, applies Translation, Rotation, then Scale

    :return: 4x4 np.array matrix
    """


    # Allow inversion of translation vector and radians which is useful if what is being represented is a "viewer" (camera) rather than an object.
    adj = 1
    if invert:
        adj = -1

    # Create Translation Matrix
    translation_matrix = mm.translationMatrix(adj * orientation_args.translation_vector.x,
                                              adj * orientation_args.translation_vector.y,
                                              adj * orientation_args.translation_vector.z)

    # Create Rotation Matrix
    rotation_matrix = mm.rotateAroundAxis(orientation_args.rotation_point.as_tuple(),
                                          orientation_args.rotation_axis.as_tuple(),
                                          radians=orientation_args.rotation_rads * adj)

    # Create Scale Matrix
    # scale_matrix = mm.scaleAroundPoint(orientation_args.scale_point.as_tuple(),
    #                                    orientation_args.scale_vector.as_tuple())
    scale_matrix = mm.scaleAroundPoint((0, 0, 0),
                                       orientation_args.scale_vector.as_tuple())

    scale_translate_matrix = mm.translationMatrix(*orientation_args.scale_adjustment.as_tuple())

    ret = scale_translate_matrix.dot(scale_matrix.dot(rotation_matrix.dot(translation_matrix)))

    return ret


def quaternion_rotation_matrix(Q):
    # https://automaticaddison.com/how-to-convert-a-quaternion-to-a-rotation-matrix/
    # https://stackoverflow.com/questions/1556260/convert-quaternion-rotation-to-rotation-matrix
    """
    Covert a quaternion into a full three-dimensional rotation matrix.

    Input
    :param Q: A 4 element array representing the quaternion (q0,q1,q2,q3)

    Output
    :return: A 4x4 element matrix representing the full 3D rotation matrix.
             This rotation matrix converts a point in the local reference
             frame to a point in the global reference frame.
    """

    # qx = Q[0]
    # qy = Q[1]
    # qz = Q[2]
    # qw = Q[3]
    #
    # ret = np.array([
    #     [1.0 - 2.0 * qy * qy - 2.0 * qz * qz, 2.0 * qx * qy - 2.0 * qz * qw, 2.0 * qx * qz + 2.0 * qy * qw, 0.0],
    #     [2.0 * qx * qy + 2.0 * qz * qw, 1.0 - 2.0 * qx * qx - 2.0 * qz * qz, 2.0 * qy * qz - 2.0 * qx * qw, 0.0],
    #     [2.0 * qx * qz - 2.0 * qy * qw, 2.0 * qy * qz + 2.0 * qx * qw, 1.0 - 2.0 * qx * qx - 2.0 * qy * qy, 0.0],
    #     [0.0, 0.0, 0.0, 1.0],
    # ])
    #
    # return ret



    # Extract the values from Q
    q0 = Q[0]
    q1 = Q[1]
    q2 = Q[2]
    q3 = Q[3]

    # First row of the rotation matrix
    r00 = 2 * (q0 * q0 + q1 * q1) - 1
    r01 = 2 * (q1 * q2 - q0 * q3)
    r02 = 2 * (q1 * q3 + q0 * q2)

    # Second row of the rotation matrix
    r10 = 2 * (q1 * q2 + q0 * q3)
    r11 = 2 * (q0 * q0 + q2 * q2) - 1
    r12 = 2 * (q2 * q3 - q0 * q1)

    # Third row of the rotation matrix
    r20 = 2 * (q1 * q3 - q0 * q2)
    r21 = 2 * (q2 * q3 + q0 * q1)
    r22 = 2 * (q0 * q0 + q3 * q3) - 1

    # 4x4 rotation matrix
    rot_matrix = np.array([[r00, r01, r02, 0],
                           [r10, r11, r12, 0],
                           [r20, r21, r22, 0],
                           [0,     0,   0, 1]])

    return rot_matrix


from dataclasses import dataclass
from coopgame.particleSystem.callables import int_callable, UpdateRenderableProtocol_provider, phase_callback
from typing import Union
from coopgame.particleSystem.protocol import UpdateRenderableProtocol
from typing import List, Tuple, Callable, Optional

@dataclass(frozen=True)
class BurstArgs:
    burst_interval_ms: Union[int, int_callable]
    particles_per_burst: Union[int, int_callable]
    max_bursts: Union[int, int_callable] = None

    @property
    def BurstIntervalMS(self):
        if callable(self.burst_interval_ms):
            return self.burst_interval_ms()
        return self.burst_interval_ms

    @property
    def MaxBursts(self):
        if callable(self.max_bursts):
            return self.max_bursts()
        return self.max_bursts

    @property
    def ParticlesPerBurst(self):
        if callable(self.particles_per_burst):
            return self.particles_per_burst()
        return self.particles_per_burst

phase_end_trigger = Callable[[UpdateRenderableProtocol], bool]


@dataclass(frozen=True)
class EmitterPhase:
    emitter_provider: UpdateRenderableProtocol_provider
    phase_end_trigger: phase_end_trigger
    phase_end_callback: phase_callback = None
    phase_start_callback: phase_callback = None

    def resolve(self, previous_phase: UpdateRenderableProtocol = None) -> UpdateRenderableProtocol:
        return self.emitter_provider(previous_phase)

@dataclass(frozen=True)
class PhaseArgs:
    phases: List[EmitterPhase]
    loop: bool = False
    max_phasers: int = None

@dataclass(frozen=True)
class ParticleGenerationArgs:
    generation_interval_ms: Union[int, int_callable] = 20
    burst_args: BurstArgs = None
    phase_args: PhaseArgs = None
    max_particles: Union[int, int_callable] = None
    init_particles: int = None

    @property
    def GenerationIntervalMS(self):
        if callable(self.generation_interval_ms):
            return self.generation_interval_ms()
        return self.generation_interval_ms

    @property
    def MaxParticles(self):
        if callable(self.max_particles):
            return self.max_particles()
        return self.max_particles






from typing import List, Callable, Any
from coopstructs.geometry.vectors.vectorN import Vector2
import coopgame.pygamehelpers as help
import pygame

class SnapState:
    def __init__(self, snap_sound: pygame.mixer.Sound = None):
        self.snap_point = None
        self.snap_sound = snap_sound
        self.snappables = []

    def set_snappables(self, snappables: List[Vector2]):
        self.snappables = snappables

    def set_snap_point(self,
                       mouse_pos: Vector2,
                       snap_distance: float,
                       draw_scale_matrix = None,
                       on_snap_callback: Callable[[Vector2], Any] = None,
                       on_unsnap_callback: Callable[[Vector2], Any] = None,
                       snappable_points: List[Vector2] = None) -> Vector2:

        # update snappable list
        if snappable_points:
            self.set_snappables(snappable_points)

        # return if no eligible snappables
        if self.snappables is None or len(self.snappables) == 0:
            return None

        # get the scaled list of points
        normal_array = help.normal_points_to_scaled_points(self.snappables,
                                                                          draw_scale_matrix=draw_scale_matrix)
        normal_snap_points = [Vector2(point[0], point[1]) for point in normal_array]

        # get the closest point in snappable list
        old = self.snap_point
        closest_scaled_point = mouse_pos.closest_within_threshold(normal_snap_points, snap_distance)

        # update snappoint
        self.snap_point = help.scaled_points_to_normal_points([closest_scaled_point], draw_scale_matrix=draw_scale_matrix)[0] if closest_scaled_point else None

        # handle callbacks
        if on_unsnap_callback is not None \
            and ((self.snap_point is None and old is not None)
                or (self.snap_point != old)):
            on_unsnap_callback(old)

        if self.snap_point is not None and old != self.snap_point:
            if on_snap_callback is not None:
                on_snap_callback(self.snap_point)

            # play the snap sound
            if self.snap_sound:
                pygame.mixer.Sound.play(self.snap_sound)

        return self.snap_point



import time
from typing import Tuple, Optional
import cooptools.geometry_utils.vector_utils as vec


def accrued_s(delta_time_ms: int,
              time_scale_seconds_per_second: int = 1) -> float:
    return delta_time_ms / 1000.0 * time_scale_seconds_per_second

def updated_velo(velocity: vec.FloatVec,
                 accel: vec.FloatVec,
                 delta_time_ms: int,
                 time_scale_seconds_per_second: int = 1,
                 ) -> vec.FloatVec:
    ''' Calculate the amount of accrued velocity in delta time and apply to position'''
    accrued_sec = accrued_s(delta_time_ms=delta_time_ms, time_scale_seconds_per_second=time_scale_seconds_per_second)
    accrued_accel = vec.scale_vector_length(accel, accrued_sec)
    return vec.add_vectors([velocity, accrued_accel])


class TimeTracker:
    def __init__(self):
        self._first = time.perf_counter()
        self._now = self._first
        self._last = self._first
        self._delta = None
        self._n_updates = 0

    def update(self, perf: int = None):
        self._last = self._now
        self._n_updates += 1
        self._now = perf if perf else time.perf_counter()
        self._delta = self._now - self._last

    def adjusted_delta_ms(self, time_scale_seconds_per_second: int = 1):
        return self.Delta_S * time_scale_seconds_per_second * 1000

    def accrued_s(self, time_scale_seconds_per_second: int = 1):
        return self.Duration_S * time_scale_seconds_per_second * 1000

    @property
    def Now(self):
        return self._now

    @property
    def Last(self):
        return self._last

    @property
    def Duration_S(self):
        return self._now - self._first

    @property
    def Delta_S(self):
        return self._delta

    @property
    def Delta_MS(self):
        return self.Delta_S * 1000

    @property
    def Avg_Update_MS(self):
        return (self._now - self._first) / self._n_updates * 1000

class Velocity:
    def __init__(self,
                 initial_m_s_vec: vec.FloatVec,
                 max_m_s: float = None,
                 ):
        self._max_m_per_s = max_m_s
        self._current_m_s_vec = initial_m_s_vec

    def set(self,
            m_s_vec: vec.FloatVec = None,
            delta_m_s_vec: vec.FloatVec = None):
        if m_s_vec is not None:
            self._current_m_s_vec = m_s_vec

        if delta_m_s_vec is not None:
            self._current_m_s_vec = vec.add_vectors([self._current_m_s_vec, delta_m_s_vec])

        self._verify()

    @staticmethod
    def _time_scale_of_vector(vector: vec.FloatVec, delta_time_ms: int, time_scale_seconds_per_second: float):
        return vec.scale_vector_length(vector, delta_time_ms / 1000.0 * time_scale_seconds_per_second)

    def _verify(self):
        if self._max_m_per_s is not None and vec.vector_len(self._current_m_s_vec) > self._max_m_per_s:
            self._current_m_s_vec = vec.scaled_to_length(self._current_m_s_vec, self._max_m_per_s)

    def update(self,
               delta_time_ms: int,
               accel: vec.FloatVec = None
               ) -> Optional[vec.FloatVec]:

        if delta_time_ms is None or delta_time_ms <= 0:
            return None

        ''' Accrued velo'''
        accrued_velo = vec.scale_vector_length(self._current_m_s_vec, delta_time_ms / 1000)

        ''' Update the velo'''
        self.set(delta_m_s_vec=accel)

        return accrued_velo

    @property
    def CurrentVelocity_M_S(self) -> vec.FloatVec:
        return self._current_m_s_vec

    def __repr__(self):
        return str(self.CurrentVelocity_M_S)

class Acceleration:
    def __init__(self,
                 initial_m_s2_vec: vec.FloatVec,
                 max_m_s2: float = None,
                 ):
        self._max_m_s2 = max_m_s2
        self._current_m_s2_vec = initial_m_s2_vec

    def set(self,
            m_s2_vec:  vec.FloatVec = None,
            delta_m_s2_vec: vec.FloatVec = None):
        if m_s2_vec is not None:
            self._current_m_s2_vec = m_s2_vec

        if delta_m_s2_vec is not None:
            self._current_m_s2_vec = vec.add_vectors([self._current_m_s2_vec, delta_m_s2_vec])

        self._verify()

    @staticmethod
    def _time_scale_of_vector(vector: vec.FloatVec, delta_time_ms: int, time_scale_seconds_per_second: float):
        return vec.scale_vector_length(vector, delta_time_ms / 1000.0 * time_scale_seconds_per_second)

    def _verify(self):
        if self._max_m_s2 is not None and vec.vector_len(self._current_m_s2_vec) > self._max_m_s2:
            self._current_m_s2_vec = vec.scaled_to_length(self._current_m_s2_vec, self._max_m_s2)

    def update(self,
               delta_time_ms: int
               ) -> Optional[vec.FloatVec]:

        if delta_time_ms is None or delta_time_ms <= 0:
            return None

        return vec.scale_vector_length(self._current_m_s2_vec, delta_time_ms / 1000)

    @property
    def CurrentAccel_M_S(self):
        return self._current_m_s2_vec

    def __repr__(self):
        return str(self.CurrentAccel_M_S)

if __name__ =="__main__":
    v = Velocity((0, 0))
    a = Acceleration((0, 0))
    tt = TimeTracker()

    while True:
        tt.update()
        if tt.Duration_S > 5:
            a.set(m_s2_vec=(1, 1))
        if tt.Duration_S > 10:
            a.set(m_s2_vec=(0, 0))

        accel = a.update(tt.Delta_MS)
        v.update(delta_time_ms=tt.Delta_MS, accel=accel)
        print(v)
        time.sleep(0.5)




from coopbugger.monitoredclass import MonitoredClass
from coopstructs.geometry import Rectangle
from coopstructs.vectors import Vector2
from coopgame.gameTemplate import GameTemplate
import harfang as hg
from coopgame.harfang.harfangColorConverter import harfang_color_from_coopcolor
from coopgame.colors import Color
from coopgame.runtime_config import RuntimeConfig
from typing import Tuple

class HarfangGameTemplate(GameTemplate):

    def __init__(self,
                 config: RuntimeConfig = None):
        self.plus = hg.GetPlus()
        hg.LoadPlugins()
        self.keyboard = hg.GetInputSystem().GetDevice("keyboard")
        super().__init__(config)



    ######## HARFANG SPECIFIC IMPLEMENTATION OF IGameTemplate

    def _quit(self):
        self.plus.RenderUninit()

    def _init_screen(self, fullscreen):
        self.plus.RenderInit(self.screen_width, self.screen_height)

    def _handle_library_specific_events(self):
        if self.keyboard.WasPressed(hg.KeyEscape) \
                or self.plus.IsAppEnded():
            self.running = False

    def _init_frame(self):
        self.plus.Clear()

    def _render_frame(self):
        self.plus.Flip()
        self.plus.EndFrame()

    def _draw_text(self,
                   text: str,
                   offset_rect: Rectangle = None,
                   color: Color = None,
                   font_size: int = None):
        if font_size is None:
            font_size = 10

        if color is None:
            color = Color.WHITE

        if offset_rect is None:
            offset_rect = Rectangle(0, 0, 1, 1)

        hcolor = harfang_color_from_coopcolor(color)
        self.plus.Text2D(offset_rect.x, offset_rect.y, f"{text}", font_size, hcolor)

    def _draw_monitoredclass_stats(self,
                                  monitoredClass: MonitoredClass,
                                  config: RuntimeConfig):

        tracked_time = [(key, val) for key, val in monitoredClass.tracked_time.items()]
        tracked_time.sort(key=lambda x: x[1], reverse=True)

        offset_rectangle = Rectangle(config.monitoredclass_txt_offset_x,
                                     self.screen_height - config.monitoredclass_txt_font_size + config.monitoredclass_txt_offset_y,
                                     1, 1)
        y_off = offset_rectangle.y

        self._draw_text(f"RunTime Stats for {type(monitoredClass).__name__}",
                        color=Color.LEMON_CHIFFON,
                        offset_rect=offset_rectangle,
                        font_size=config.monitoredclass_txt_font_size)
        y_off -= config.monitoredclass_txt_font_size + 3
        for key, val in tracked_time:
            offset_rectangle.y = y_off
            self._draw_text(f"{key}: {round(val, 2)} sec",
                            color=Color.LEMON_CHIFFON,
                            offset_rect=offset_rectangle,
                            font_size=config.monitoredclass_txt_font_size)
            y_off -= config.monitoredclass_txt_font_size + 3


    @staticmethod
    def mouse_pos_as_vector() -> Vector2:
        # get the mouse device
        mouse_device = hg.GetInputSystem().GetDevice("mouse")
        return Vector2(mouse_device.GetValue(hg.InputAxisX), mouse_device.GetValue(hg.InputAxisY))

    ############ USER DEFINED FOR GAME ####################
    def initialize_game(self):
        raise NotImplementedError()

    def handle_left_click(self, pressed_keys):
        raise NotImplementedError()

    def handle_right_click(self, pressed_keys):
        raise NotImplementedError()

    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        raise NotImplementedError()

    def handle_mouse_scroll_up(self, pressed_keys):
        raise NotImplementedError()

    def handle_mouse_scroll_down(self, pressed_keys):
        raise NotImplementedError()

    def draw(self, frames: int, debug_mode: bool = False):
        raise NotImplementedError()

    def model_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    def sprite_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    def on_resize(self):
        raise NotImplementedError()

    def register_button(self, id, text, callback, postion_rect):
        raise NotImplementedError()

    def register_keys(self):
        raise NotImplementedError()




from enum import Enum

def add_succ_and_pred_maps(cls):
    succ_map = {}
    pred_map = {}
    first = None
    cur = None
    nxt = None
    last = None
    for val in cls.__members__.values():
        if first is None:
            first = val

        if cur is None:
            cur = val
        elif nxt is None:
            nxt = val

        last = val

        if cur is not None and nxt is not None and cur != nxt:
            succ_map[cur] = nxt
            pred_map[nxt] = cur
            cur = nxt
            nxt = None

    # add wrap arounds
    succ_map[last] = first
    pred_map[first] = last

    cls._succ_map = succ_map
    cls._pred_map = pred_map

    def succ(self):
        return self._succ_map[self]

    def pred(self):
        return self._pred_map[self]

    cls.succ = succ
    cls.pred = pred
    return cls



@add_succ_and_pred_maps
class Orientation(Enum):
    UP=1
    RIGHT=2
    DOWN=3
    LEFT=4

@add_succ_and_pred_maps
class ConnectionType(Enum):
    ARC = 1
    LINE = 2
    BEZIER = 3

@add_succ_and_pred_maps
class EntryType(Enum):
    CRAWLER = 1
    GRIDCLICK = 2

@add_succ_and_pred_maps
class GridDrawType(Enum):
    BOXES = 1
    LINES = 2


from coopgame.renderedObjectHandling.wireframes import Wireframe
from coopgame.renderedObjectHandling.objectOrientationArgs import ObjectOrientationArgs
from coopgame.coordinateConverter import CoordinateConverter as cc
from coopgame.renderedObjectHandling.objectOrientationMatrixFactory import object_orientation_matrix
from coopgame.renderedObjectHandling.renderedViewArgs import RenderedViewArgs
import math
import coopgame.renderedObjectHandling.MatrixManipulation as mm
import numpy as np
from coopgame.models.primitives.coopquaternion import CoopQuaternion

class RenderedObject:
    def __init__(self,
                 wireframe: Wireframe):
        self._object_wireframe = wireframe
        self.orientation = ObjectOrientationArgs()

    @property
    def world_nodes(self):
        centre = self._object_wireframe.object_centre[:3] + self.orientation.translation_vector.as_tuple()
        return self._apply_orientation(points=self._object_wireframe.nodes,
                                       orientation_args=self.orientation,
                                       centre=centre)

        # # Apply translation
        # translated_nodes = cc.apply_matrix_to_array(self._object_wireframe.nodes, self._translation_matrix(self.orientation))
        #
        # # Apply Quaternion rotation
        # centre = self._object_wireframe.object_centre[:3] + self.orientation.translation_vector.as_tuple()
        # rotated_points = self.orientation.quaternion.rotate_all(translated_nodes, centre)
        #
        # # Scale the points
        # scaled_nodes = cc.apply_matrix_to_array(rotated_points, self._scale_matrix(self.orientation))
        #
        # return scaled_nodes


    @property
    def world_edges(self):
        return self._object_wireframe.translated_edges(object_orientation_matrix(self.orientation))

    @property
    def world_center(self):
        ret = self._object_wireframe.findCentre(object_orientation_matrix(self.orientation))

        if not math.isclose(ret[0], self._object_wireframe.nodes[0][0] + self._object_wireframe.size / 2 + self.orientation.translation_vector.x):
            print(ValueError("Error calculating the world_center"))

        return ret

    def camera_nodes(self, render_view_args: RenderedViewArgs):
        centre = self.world_center
        return self._apply_orientation(self.world_nodes,
                                       render_view_args.camera_orientation,
                                       centre=centre,
                                       invert=True)
        # return cc.apply_matrix_to_array(self.world_nodes,
        #                                 object_orientation_matrix(render_view_args.camera_orientation,
        #                                                           invert=True))

    def camera_edges(self, render_view_args: RenderedViewArgs):
        cam_nodes = self.camera_nodes(render_view_args)
        return [(cam_nodes[n1][:2], cam_nodes[n2][:2]) for n1, n2 in self._object_wireframe.edge_node_ids]

    def camera_center(self, render_view_args: RenderedViewArgs):
        centre = self.world_center
        return self._apply_orientation(self.world_center,
                                       render_view_args.camera_orientation,
                                       centre=centre,
                                       invert=True)
        # return cc.apply_matrix_to_array(self.world_center,
        #                                 object_orientation_matrix(render_view_args.camera_orientation,
        #                                                           invert=True))

    @staticmethod
    def _translation_matrix(orientation_args: ObjectOrientationArgs, invert:bool = False):
        # Allow inversion of translation vector and radians which is useful if what is being represented is a "viewer" (camera) rather than an object.
        adj = 1
        if invert:
            adj = -1

        # Create Translation Matrix
        translation_matrix = mm.translationMatrix(adj * orientation_args.translation_vector.x,
                                                  adj * orientation_args.translation_vector.y,
                                                  adj * orientation_args.translation_vector.z)

        return translation_matrix

    @staticmethod
    def _scale_matrix(orientation_args: ObjectOrientationArgs):
        # Create Scale Matrix
        scale_matrix = mm.scaleAroundPoint(orientation_args.scale_point.as_tuple(),
                                           orientation_args.scale_vector.as_tuple())

        return scale_matrix


    def _apply_orientation(self, points: np.array, orientation_args: ObjectOrientationArgs, centre: np.array, invert:bool=False):
        # Apply translation
        translated_nodes = cc.apply_matrix_to_array(points, self._translation_matrix(orientation_args,
                                                                                     invert=invert))

        # Apply Quaternion rotation
        if invert:
            axis = orientation_args.quaternion.axis if orientation_args.quaternion.angle != 0 else [1, 0, 0]
            quat = CoopQuaternion(axis=axis, angle=-orientation_args.quaternion.angle)
        else:
            quat = orientation_args.quaternion
        rotated_points = quat.rotate_all(translated_nodes, centre)

        # Scale the points
        scaled_nodes = cc.apply_matrix_to_array(rotated_points, self._scale_matrix(orientation_args))

        return scaled_nodes
